from .model import GroupModel, GroupCollection, GroupSchema
from ..app import App


def get_group(request, identifier):
    storage = request.app.root.get_authnz_provider(
    ).get_authn_storage(request, GroupSchema)
    return storage.get(identifier)


def get_group_collection(request):
    return GroupCollection(request,
                           request.app.root.get_authnz_provider().get_authn_storage(request, GroupSchema))


@App.path(model=GroupModel,
          path='group/{identifier}')
def _get_group(app, request, identifier):
    return get_group(request, identifier)


@App.path(model=GroupCollection,
          path='group')
def _get_group_collection(app, request):
    return get_group_collection(request)



import jsonobject
from morpfw.crud import Schema
from morpfw.crud.validator import regex_validator
from ..model import NAME_PATTERN, EMAIL_PATTERN
from ..app import App


class RegistrationSchema(jsonobject.JsonObject):
    # pattern=NAME_PATTERN)
    username = jsonobject.StringProperty(
        required=True, validators=[regex_validator(NAME_PATTERN, 'name')])
    # pattern=EMAIL_PATTERN
    email = jsonobject.StringProperty(required=True, validators=[
                                      regex_validator(EMAIL_PATTERN, 'email')])
    password = jsonobject.StringProperty(required=True)
    password_validate = jsonobject.StringProperty(required=True, default='')


class LoginSchema(jsonobject.JsonObject):

    username = jsonobject.StringProperty(required=True)
    password = jsonobject.StringProperty(required=True)


class UserSchema(Schema):

    username = jsonobject.StringProperty(
        required=True, validators=[regex_validator(NAME_PATTERN, 'name')])  # , pattern=NAME_PATTERN)
    # , pattern=EMAIL_PATTERN)
    email = jsonobject.StringProperty(required=True, validators=[
                                      regex_validator(EMAIL_PATTERN, 'email')])
    password = jsonobject.StringProperty(required=False)
    groups = jsonobject.ListProperty(
        str, required=False)  # pattern=NAME_PATTERN
    attrs = jsonobject.DictProperty(required=False)
    state = jsonobject.StringProperty(required=False)
    created = jsonobject.StringProperty(required=False)
    modified = jsonobject.StringProperty(required=False)
    nonce = jsonobject.StringProperty(required=False)


@App.identifierfields(schema=UserSchema)
def user_identifierfields(schema):
    return ['username']



from .crud_common import App as BaseApp
from morpfw.crud.model import Collection, Model
from morpfw.crud.schema import Schema
from morpfw.crud.storage.elasticsearchstorage import ElasticSearchStorage
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from .crud_common import get_client, run_jslcrud_test, PageCollection, PageModel
from .crud_common import NamedObjectCollection, NamedObjectModel
from .crud_common import BlobObjectCollection, BlobObjectModel
from .crud_common import FSBLOB_DIR
import pprint
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from more.basicauth import BasicAuthIdentityPolicy
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session
import morpfw.crud.signals as signals
from elasticsearch import Elasticsearch
import jsonobject

Session = sessionmaker()
register_session(Session)


class ESClientRequest(Request):

    @reify
    def es_client(self):
        return Elasticsearch(hosts='127.0.0.1:9085')


class App(BaseApp, TransactionApp):

    request_class = ESClientRequest


class PageStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'page'
    refresh = 'wait_for'
    model = PageModel


@App.path(model=PageCollection, path='pages')
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='pages/{identifier}')
def model_factory(request, identifier):
    storage = PageStorage(request)
    return storage.get(identifier)


class ObjectSchema(Schema):

    id = jsonobject.StringProperty(required=False)
    uuid = jsonobject.StringProperty(required=False)
    body = jsonobject.StringProperty(required=True, default='')
    created_flag = jsonobject.BooleanProperty(required=False, default=False)
    updated_flag = jsonobject.BooleanProperty(required=False, default=False)


@App.identifierfields(schema=ObjectSchema)
def object_identifierfields(schema):
    return ['id']


@App.default_identifier(schema=ObjectSchema)
def object_default_identifier(schema, obj, request):
    return None


class ObjectModel(Model):
    schema = ObjectSchema


class ObjectCollection(Collection):
    schema = ObjectSchema


@App.subscribe(signal=signals.OBJECT_CREATED, model=ObjectModel)
def object_created(app, request, obj, signal):
    obj.data['created_flag'] = True


@App.subscribe(signal=signals.OBJECT_UPDATED, model=ObjectModel)
def object_updated(app, request, obj, signal):
    obj.data['updated_flag'] = True


class ObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'object'
    refresh = 'wait_for'
    auto_id = True
    model = ObjectModel


@App.json(model=ObjectCollection, name='get_uuid')
def get_object_by_uuid(context, request):
    uuid = request.GET.get('uuid')
    return context.get_by_uuid(uuid).json()


@App.path(model=ObjectCollection, path='objects')
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path='objects/{identifier}')
def object_model_factory(request, identifier):
    storage = ObjectStorage(request)
    return storage.get(identifier)


class NamedObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'namedobject'
    refresh = 'wait_for'
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path='named_objects')
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path='named_objects/{identifier}')
def namedobject_model_factory(request, identifier):
    storage = NamedObjectStorage(request)
    return storage.get(identifier)


class BlobObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'blobobject'
    refresh = 'wait_for'
    model = BlobObjectModel


@App.path(model=BlobObjectCollection, path='blob_objects')
def blobobject_collection_factory(request):
    blobstorage = FSBlobStorage(FSBLOB_DIR)
    storage = BlobObjectStorage(request, blobstorage=blobstorage)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path='blob_objects/{identifier}')
def blobobject_model_factory(request, identifier):
    blobstorage = FSBlobStorage(FSBLOB_DIR)
    storage = BlobObjectStorage(request, blobstorage)
    return storage.get(identifier)


def test_elasticsearchstorage(es_client):
    es_client.indices.create('test-index', body={
        'settings': {
            'number_of_shards': 1,
            'number_of_replicas': 0
        }
    })
    es_client.transport.perform_request('PUT', '/test-index/_mapping/page',
                                        body={
                                            'properties': {
                                                'title': {
                                                    'type': 'text',
                                                    'fielddata': True
                                                }
                                            }
                                        })
    client = get_client(App)
    run_jslcrud_test(client)



import copy
from ..errors import BlobStorageNotImplementedError


class BaseStorage(object):

    use_transactions = True

    blobstorage = None

    @property
    def model(self):
        raise NotImplementedError

    def __init__(self, request, blobstorage=None):
        self.request = request
        self.app = request.app
        self.blobstorage = blobstorage

    def set_identifier(self, obj, identifier):
        for f, v in zip(
                self.app.get_identifierfields(self.model.schema),
                identifier.split(
                    self.app.get_compositekey_separator())):
            obj[f] = v

    def set_schema_defaults(self, data):
        obj = self.model.schema(data)
        return obj.to_json()

    def create(self, data):
        raise NotImplementedError

    def search(self, query=None, limit=None):
        raise NotImplementedError

    def get(self, identifier):
        raise NotImplementedError

    def get_by_uuid(self, uuid):
        raise NotImplementedError

    def update(self, identifier, data):
        raise NotImplementedError

    def delete(self, identifier, model):
        raise NotImplementedError

    def _blob_guard(self):
        if self.blobstorage is None:
            raise BlobStorageNotImplementedError(
                'Storage does not implement blobstorage')

    def get_blob(self, uuid):
        self._blob_guard()
        return self.blobstorage.get(uuid)

    def put_blob(self, fileobj, filename, mimetype=None, size=None, encoding=None):
        self._blob_guard()
        blob = self.blobstorage.put(
            fileobj, filename, mimetype, size, encoding)
        return blob

    def delete_blob(self, uuid):
        self._blob_guard()
        self.blobstorage.delete(uuid)



from .jsonprovider import JsonObjectProvider
from ..app import App
from ..storage.elasticsearchstorage import ElasticSearchStorage
from ..model import Model
import jsonobject


@App.dataprovider(schema=jsonobject.JsonObject, obj=jsonobject.JsonObject,
                  storage=ElasticSearchStorage)
def get_dataprovider(schema, obj,  storage):
    return JsonObjectProvider(schema, obj, storage)



import morepath
from more.jwtauth import JWTIdentityPolicy
from morpfw.crud import App as CRUDApp
import reg
import dectate
from . import action
from morepath.reify import reify
import json
from typing import List, Optional, Type
import morpfw

_REGISTERED_APPS: List[morepath.App] = []


class App(CRUDApp):

    authn_storage = dectate.directive(action.StorageAction)

    def get_authn_storage(self, request: morepath.Request, schema: Type[morpfw.Schema]):
        name = request.app.settings.application.authn_storage
        storage_opts = request.app.settings.application.authn_storage_opts
        return self._get_authn_storage(name, schema)(request=request,
                                                     **storage_opts)

    @reg.dispatch_method(reg.match_key('name',
                                       lambda self, name, schema: name),
                         reg.match_class('schema',
                                         lambda self, name, schema: schema))
    def _get_authn_storage(self, name, schema):
        raise NotImplementedError

    def get_roles(self, request: morepath.Request, username: Optional[str] = None,
                  groupname: str = '__default__'):
        if username is None:
            username = request.identity.userid
        from .group.model import GroupSchema
        storage = self.get_authn_storage(request, GroupSchema)
        return storage.get_group_user_roles(groupname, username)

    def has_role(self, request: morepath.Request, rolename: str,
                 username: Optional[str] = None,
                 groupname: str = '__default__'):
        return rolename in self.get_roles(request, username, groupname)

    def __repr__(self):
        return u'AuthManager'



from more.jwtauth import JWTIdentityPolicy
from morepath import Identity, NO_IDENTITY
import rulez
from ..path import get_apikey_collection
from .base import AuthnPolicy as BaseAuthnPolicy


class JWTWithAPIKeyIdentityPolicy(JWTIdentityPolicy):

    def identify(self, request):
        api_key = request.headers.get('X-API-KEY', None)
        if api_key:
            api_identity, api_secret = api_key.split('.')
            apikeys = get_apikey_collection(request)
            keys = apikeys.search(
                rulez.field['api_identity'] == api_identity, secure=False)
            if keys and keys[0].data['api_secret'] == api_secret:
                username = keys[0].data['username']
                return Identity(userid=username)
        return super(JWTWithAPIKeyIdentityPolicy, self).identify(request)


class AuthnPolicy(BaseAuthnPolicy):

    @classmethod
    def get_identity_policy(cls, settings):
        jwtauth_settings = settings.jwtauth
        if jwtauth_settings:
            # Pass the settings dictionary to the identity policy.
            return JWTWithAPIKeyIdentityPolicy(**jwtauth_settings.__dict__.copy())
        raise Exception('JWTAuth configuration is not set')

    @classmethod
    def verify_identity(cls, app, identity):
        return True





def rellink(context, request, name=None, method='GET', link_name=None):
    link_name = link_name or name
    if name:
        name = '+%s' % name
    if name is None:
        link_name = 'self'
    return {
        'rel': link_name,
        'type': method,
        'href': request.link(context, name)
    }




class AuthnPolicy(object):

    @classmethod
    def get_identity_policy(cls, settings):
        raise NotImplementedError

    @classmethod
    def verify_identity(cls, app, identity):
        raise NotImplementedError



from ..app import App
from .model import UserCollection, UserModel
from ..path import get_user_collection
from .model import UserSchema, LoginSchema
from .model import RegistrationSchema
from ..exc import UserExistsError
import morepath
from ..validator import validate
from ..utils import rellink
from .. import permission
from webob.exc import HTTPNotFound
from more.jwtauth import (
    verify_refresh_request, InvalidTokenError, ExpiredSignatureError
)
from morpfw.crud.util import jsonobject_to_jsl
from morpfw.crud.validator import validate_schema
from morpfw.crud.errors import AlreadyExistsError


@App.json(model=UserCollection, name='register', request_method='POST',
          permission=permission.Register, load=validate_schema(schema=RegistrationSchema))
def register(context, request, load):
    """Validate the username and password and create the user."""
    data = request.json
    res = validate(data, jsonobject_to_jsl(
        RegistrationSchema).get_schema())
    if res:
        @request.after
        def set_error(response):
            response.status = 422
        return {
            'status': 'error',
            'field_errors': [{'message': res[x]} for x in res.keys()]
        }

    if data['password'] != data['password_validate']:
        @request.after
        def adjust_response(response):
            response.status = 422

        return {'status': 'error',
                'message': 'Password confirmation does not match'}

    if 'state' not in data.keys() or not data['state']:
        data['state'] = request.app.settings.application.new_user_state
    obj = context.create(data)
    return {'status': 'success'}


@App.json(model=UserCollection, name='login')
def login(context, request):
    return {
        'schema': jsonobject_to_jsl(LoginSchema).get_schema(),
        'links': [rellink(context, request, 'login', 'POST')]
    }


@App.json(model=UserCollection, name='create', request_method='POST')
def reject_create(context, request):
    raise HTTPNotFound()


@App.json(model=UserCollection, name='login', request_method='POST')
def process_login(context, request):
    """Authenticate username and password and log in user"""
    username = request.json['username']
    password = request.json['password']

    # Do the password validation.
    if not context.authenticate(username, password):
        @request.after
        def adjust_status(response):
            response.status = 401
        return {
            'status': 'error',
            'error': {
                'code': 401,
                'message': 'Invalid Username / Password'
            }
        }

    @request.after
    def remember(response):
        """Remember the identity of the user logged in."""
        # We pass the extra info to the identity object.
        response.headers.add('Access-Control-Expose-Headers', 'Authorization')
        identity = morepath.Identity(username)
        request.app.remember_identity(response, request, identity)

    return {
        'status': 'success'
    }


@App.json(model=UserCollection, name='refresh_token')
def refresh_token(context, request):
    try:
        # Verifies if we're allowed to refresh the token.
        # In this case returns the userid.
        # If not raises exceptions based on InvalidTokenError.
        # If expired this is a ExpiredSignatureError.
        username = verify_refresh_request(request)
    except ExpiredSignatureError:
        @request.after
        def expired_nonce_or_token(response):
            response.status_code = 403
        return {
            'status': 'error',
            'message': 'Your session has expired.'
        }
    except InvalidTokenError as e:
        @request.after
        def invalid_token(response):
            response.status_code = 403
        return {
            'status': 'error',
            'message': 'Could not refresh your token. %s' % str(e)
        }
    else:
        @request.after
        def remember(response):
            # create the identity with the userid and updated user info
            identity = morepath.Identity(username)
            # create the updated token and set it in the response header
            request.app.remember_identity(response, request, identity)

        return {
            'status': 'success',
            'message': 'Token successfully refreshed'
        }


@App.json(model=UserCollection, name='logout')
def logout(context, request):
    """Log out the user."""

    @request.after
    def forget(response):
        request.app.forget_identity(response, request)

    return {
        'status': 'success'
    }


@App.json(model=UserModel, request_method='PATCH')
def update(context, request):
    error = None
    if 'password' in request.json.keys():
        error = {
            'status': 'error',
            'message': 'Changing password is not allowed through this method'
        }
    elif 'username' in request.json.keys():
        error = {
            'status': 'error',
            'message': 'Changing username is not allowed'
        }

    if error:
        @request.after
        def adjust_status(response):
            response.status = 422

        return error

    context.update(request.json)
    return {'status': 'success'}


@App.json(model=UserModel, name='roles')
def roles(context, request):
    return context.group_roles()


@App.json(model=UserModel, name='change_password', request_method='POST')
def change_password(context, request):
    data = request.json
    error = None
    current_password = data.get('password', '')
    if not request.app.has_role(request, 'administrator'):
        if not context.validate(current_password, check_state=False):
            error = 'Invalid password'

    if not error and data['new_password'] != data['new_password_validate']:
        error = 'Password confirmation does not match'

    if error:
        @request.after
        def adjust_status(response):
            response.status = 422
        return {
            'status': 'error',
            'message': error
        }

    context.change_password(current_password, data['new_password'])
    return {'status': 'success'}


@App.json(model=UserModel, request_method='DELETE')
def delete(context, request):
    sm = context.state_machine()
    try:
        sm.delete()
    except AttributeError:
        @request.after
        def adjust_status(response):
            response.status = 404

        return {
            'status': 'error',
            'message': 'User no longer exists'
        }
    return {'status': 'success'}



import sqlalchemy as sa
import uuid
from sqlalchemy.ext.declarative import declarative_base
from rulez import compile_condition
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy import func
from datetime import datetime
from decimal import Decimal
from .base import BaseStorage
from ..app import App
import jsl
import uuid


class SQLStorage(BaseStorage):

    _temp = {}

    @property
    def orm_model(self):
        raise NotImplementedError

    @property
    def session(self):
        return self.request.db_session

    def create(self, data):
        o = self.orm_model()
        dst = self.app.get_dataprovider(self.model.schema, o, self)
        src = self.app.get_dataprovider(self.model.schema, data, self)
        for k, v in src.items():
            dst[k] = v
        m = self.model(self.request, self, o)
        identifier = m.identifier
        self.session.add(o)
        self.session.flush()
        self.session.refresh(o)
        return m

    def aggregate(self, query=None, group=None, order_by=None):
        group_bys = []

        if group:
            fields = []
            for k, v in group.items():
                if isinstance(v, str):
                    c = getattr(self.orm_model, v)
                    fields.append(c)
                    group_bys.append(c)
                elif isinstance(v, dict):
                    ff = v['function']
                    f = v['field']
                    c = getattr(self.orm_model, f)
                    if ff == 'count':
                        fields.append(func.count(c).label(k))
                    elif ff == 'sum':
                        fields.append(func.sum(c).label(k))
                    elif ff == 'avg':
                        fields.append(func.avg(c).label(k))
                    elif ff == 'year':
                        op = func.date_part('YEAR', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    elif ff == 'month':
                        op = func.date_part('MONTH', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    elif ff == 'day':
                        op = func.date_part('DAY', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    else:
                        raise ValueError('Unknown function %s' % ff)
        else:
            fields = [self.orm_model]

        if query:
            f = compile_condition('sqlalchemy', query)
            filterquery = f(self.orm_model)
            filterquery = sa.and_(
                self.orm_model.deleted.is_(None), filterquery)
            q = self.session.query(*fields).filter(filterquery)
        else:
            q = self.session.query(
                *fields).filter(self.orm_model.deleted.is_(None))

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ['asc', 'desc']:
                raise KeyError(d)
            colattr = getattr(self.orm_model, col)
            if d == 'desc':
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)

        if group_bys:
            q = q.group_by(*group_bys)

        results = []
        for o in q.all():
            d = o._asdict()
            for k, v in d.items():
                if isinstance(v, datetime):
                    d[k] = v.isoformat()
                elif isinstance(v, Decimal):
                    d[k] = float(v)
                elif isinstance(v, uuid.UUID):
                    d[k] = v.hex

            results.append(d)
        return results

    def search(self, query=None, offset=None, limit=None, order_by=None):
        if query:
            f = compile_condition('sqlalchemy', query)
            filterquery = f(self.orm_model)
            q = self.session.query(self.orm_model).filter(sa.and_(
                self.orm_model.deleted.is_(None), filterquery))
        else:
            q = self.session.query(self.orm_model).filter(
                self.orm_model.deleted.is_(None))

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ['asc', 'desc']:
                raise KeyError(d)
            colattr = getattr(self.orm_model, col)
            if d == 'desc':
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)
        if offset is not None:
            q = q.offset(offset)
        if limit is not None:
            q = q.limit(limit)

        return [self.model(self.request, self, o) for o in q.all()]

    def get(self, identifier):
        qs = []
        for f, v in zip(
                self.app.get_identifierfields(self.model.schema),
                identifier.split(self.app.get_compositekey_separator())):
            qs.append(getattr(self.orm_model, f) == v)
        q = self.session.query(self.orm_model).filter(
            sa.and_(self.orm_model.deleted.is_(None), *qs))
        r = q.first()
        if not r:
            return None
        return self.model(self.request, self, r)

    def get_by_uuid(self, uuid):
        uuid_field = self.app.get_uuidfield(self.model.schema)
        if getattr(self.orm_model, uuid_field, None) is None:
            raise AttributeError('%s does not have %s field' %
                                 (self.orm_model, uuid_field))
        qs = getattr(self.orm_model, uuid_field) == uuid
        q = self.session.query(self.orm_model).filter(qs)
        r = q.first()
        if not r:
            return None
        return self.model(self.request, self, r)

    def update(self, identifier, data):
        qs = []
        for f, v in zip(
                self.app.get_identifierfields(self.model.schema),
                identifier.split(
                    self.app.get_compositekey_separator())):
            qs.append(getattr(self.orm_model, f) == v)
        qs.append(self.orm_model.deleted.is_(None))
        q = self.session.query(self.orm_model).filter(sa.and_(*qs))

        r = q.first()
        if not r:
            raise ValueError(identifier)

        d = self.app.get_dataprovider(self.model.schema, r, self)
        for k, v in data.items():
            if d.get(k, None) != v:
                d[k] = v

        return self.model(self.request, self, r)

    def delete(self, identifier, model):
        model['deleted'] = datetime.utcnow()


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class BaseMixin(object):

    uuid = sa.Column(GUID, primary_key=True, default=uuid.uuid4)
    created = sa.Column(sa.DateTime, default=datetime.utcnow)
    creator = sa.Column(sa.String(length=1024))
    modified = sa.Column(sa.DateTime, default=datetime.utcnow)
    state = sa.Column(sa.String(length=1024))
    deleted = sa.Column(sa.DateTime)


Base = declarative_base(cls=BaseMixin)



from .storage import Group, Membership, RoleAssignment
from ..model import NAME_PATTERN
from morpfw.crud import Collection, Model, Schema
from ..app import App
import jsl
from morpfw.crud import errors as cruderrors
from .. import exc
import jsonobject
from .schema import GroupSchema, MemberSchema
from ..exc import GroupExistsError


class GroupCollection(Collection):
    schema = GroupSchema
    exist_exc = GroupExistsError


class GroupModel(Model):
    schema = GroupSchema

    def members(self):
        members = self.storage.get_members(self.identifier)
        active_members = [
            member for member in members if member.data.get('state') == 'active']
        return active_members

    def add_members(self, usernames):
        self.storage.add_group_members(self.identifier, usernames)

    def remove_members(self, usernames):
        self.storage.remove_group_members(self.identifier, usernames)

    def get_member_roles(self, username):
        return self.storage.get_group_user_roles(self.identifier, username)

    def grant_member_role(self, username, rolename):
        if username not in [m.identifier for m in self.members()]:
            self.add_members([username])
        self.storage.grant_group_user_role(self.identifier, username, rolename)

    def revoke_member_role(self, username, rolename):
        self.storage.revoke_group_user_role(
            self.identifier, username, rolename)
        if (not self.get_member_roles(username) and
                self.identifier != '__default__'):
            self.remove_members([username])



from ..app import App
from ..user.model import User, UserSchema, UserModel
from ..group.model import Group, GroupSchema, GroupModel
from ..apikey.model import APIKeyModel, APIKeySchema, APIKey
from morpfw.crud.storage.sqlstorage import SQLStorage
from morpfw.crud import errors as cruderrors
from .. import dbmodel as db
from .interfaces import IStorage
import hashlib
import sqlalchemy as sa
from .. import exc


def hash(data):
    return hashlib.sha256(data.encode('utf-8')).hexdigest()


class UserSQLStorage(SQLStorage):
    model = UserModel
    orm_model = User

    def create(self, data):
        data['password'] = hash(data['password'])
        return super(UserSQLStorage, self).create(data)

    def get_by_email(self, email):
        q = self.session.query(db.User).filter(db.User.email == email)
        u = q.first()
        if not u:
            return None
        return self.model(self.request, self, u)

    def change_password(self, username, new_password):
        q = self.session.query(db.User).filter(db.User.username == username)
        u = q.first()
        if not u:
            raise ValueError("Unknown User %s" % username)
        u.password = hash(new_password)

    def get_user_groups(self, username):
        q = self.session.query(db.User).filter(db.User.username == username)
        u = q.first()
        if not u:
            raise ValueError("Unknown user %s" % username)
        q = self.session.query(db.Membership).filter(
            db.Membership.user_id == u.uuid)
        membership = q.all()
        groupstorage = self.request.app.get_authn_storage(
            self.request, GroupSchema)
        res = [groupstorage.get_by_uuid(m.group_id)
               for m in membership]
        return res

    def validate(self, username, password):

        q = self.session.query(db.User).filter(db.User.username == username)
        u = q.first()
        if not u:
            raise ValueError("Unknown User %s" % username)
        return u.password == hash(password)


class APIKeySQLStorage(SQLStorage):
    model = APIKeyModel
    orm_model = APIKey


class GroupSQLStorage(SQLStorage):
    model = GroupModel
    orm_model = Group

    def get_members(self, groupname):
        q = (self.session.query(db.User)
             .join(db.Membership)
             .join(db.Group)
             .filter(db.Group.groupname == groupname))
        members = []
        user_storage = UserSQLStorage(self.request)
        for m in q.all():
            members.append(UserModel(self.request, user_storage, m))
        return members

    def add_group_members(self, groupname, usernames):
        # FIXME: not using sqlalchemy relations might impact performance
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise ValueError("Group Does Not Exist %s" % groupname)
        gid = g.uuid
        for username in usernames:
            u = self.session.query(db.User).filter(
                db.User.username == username).first()
            if not u:
                raise ValueError("User Does Not Exist %s" % username)
            uid = u.uuid
            e = self.session.query(db.Membership).filter(
                sa.and_(db.Membership.group_id == gid,
                        db.Membership.user_id == uid)).first()
            if not e:
                m = db.Membership()
                m.group_id = gid
                m.user_id = uid
                self.session.add(m)

    def remove_group_members(self, groupname, usernames):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.uuid
        for username in usernames:
            u = self.session.query(db.User).filter(
                db.User.username == username).first()
            if not u:
                raise exc.UserDoesNotExistsError(username)
            uid = u.uuid
            members = self.session.query(db.Membership).filter(
                sa.and_(db.Membership.group_id == gid,
                        db.Membership.user_id == uid)).all()
            for m in members:
                self.session.delete(m)

    def get_group_user_roles(self, groupname, username):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.uuid
        u = self.session.query(db.User).filter(
            db.User.username == username).first()
        if not u:
            raise exc.UserDoesNotExistsError(username)
        uid = u.uuid
        roles = (self.session.query(db.RoleAssignment)
                 .join(db.Membership)
                 .filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).all())
        return [r.rolename for r in roles]

    def grant_group_user_role(self, groupname, username, rolename):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.uuid
        u = self.session.query(db.User).filter(
            db.User.username == username).first()
        if not u:
            raise exc.UserDoesNotExistsError(username)
        uid = u.uuid
        m = self.session.query(db.Membership).filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).first()
        if not m:
            raise exc.MembershipError(username, groupname)
        ra = self.session.query(db.RoleAssignment).filter(
            db.RoleAssignment.membership_id == m.uuid,
            db.RoleAssignment.rolename == rolename).first()
        if ra:
            return
        r = db.RoleAssignment()
        r.membership_id = m.uuid
        r.rolename = rolename
        self.session.add(r)

    def revoke_group_user_role(self, groupname, username, rolename):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.uuid
        u = self.session.query(db.User).filter(
            db.User.username == username).first()
        if not u:
            raise exc.UserDoesNotExistsError(username)
        uid = u.uuid
        m = self.session.query(db.Membership).filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).first()
        if not m:
            raise exc.MembershipError(username, groupname)
        ra = self.session.query(db.RoleAssignment).filter(
            db.RoleAssignment.membership_id == m.uuid,
            db.RoleAssignment.rolename == rolename).first()
        if ra:
            self.session.delete(ra)


def userstorage_factory(request, *args, **kwargs):
    return UserSQLStorage(request)


def apikeystorage_factory(request, *args, **kwargs):
    return APIKeySQLStorage(request)


@App.authn_storage('sqlstorage', APIKeySchema)
def get_apikeystorage_factory(name, schema):
    return apikeystorage_factory


@App.authn_storage('sqlstorage', UserSchema)
def get_userstorage_factory(name, schema):
    return userstorage_factory


def groupstorage_factory(request, *args, **kwargs):
    return GroupSQLStorage(request)


@App.authn_storage('sqlstorage', GroupSchema)
def get_groupstorage_factory(name, schema):
    return groupstorage_factory



import importlib
import reg
import morepath
from .app import SQLApp, Session, BaseApp
from .sql import Base
import os
from zope.sqlalchemy import register as register_session
from more.basicauth import BasicAuthIdentityPolicy
from .exc import ConfigurationError
import transaction
import sqlalchemy
from celery import Celery
import yaml
import copy

default_settings = yaml.load(
    open(os.path.join(os.path.dirname(__file__), 'default_settings.yml')))


@reg.dispatch(reg.match_class('app', lambda app, *args, **kwargs: app))
def create_app(app, settings, scan=True, **kwargs):
    raise NotImplementedError


@create_app.register(app=BaseApp)
def create_baseapp(app, settings, scan=True, **kwargs):

    s = copy.deepcopy(default_settings)
    for k in settings.keys():
        if k in s.keys():
            for j, v in settings[k].items():
                s[k][j] = v
        else:
            s[k] = settings[k]

    settings = s

    # initialize app

    if scan:
        morepath.autoscan()

    app_settings = settings['application']

    authnpol_mod, authnpol_clsname = (
        app_settings['authn_policy'].strip().split(':'))
    authnpolicy = getattr(importlib.import_module(
        authnpol_mod), authnpol_clsname)

    authnzprv_mod, authnzprv_factory = (
        app_settings['authnz_provider'].strip().split(':'))
    authnz_provider = getattr(importlib.import_module(
        authnzprv_mod), authnzprv_factory)

    get_identity_policy = authnpolicy.get_identity_policy
    verify_identity = authnpolicy.verify_identity

    for iapp in app_settings['mounted_apps']:
        iapp_path = iapp['app']
        iapp_mod, iapp_clsname = iapp_path.strip().split(':')
        iapp_cls = getattr(importlib.import_module(iapp_mod), iapp_clsname)

        if iapp.get('authn_policy', None):
            iapp_authnpol_mod, iapp_authnpol_clsname = (
                iapp['authn_policy'].strip().split(':'))
            iapp_authnpolicy = getattr(importlib.import_module(
                iapp_authnpol_mod), iapp_authnpol_clsname)
            iapp_get_identity_policy = iapp_authnpolicy.get_identity_policy
            iapp_verify_identity = iapp_authnpolicy.verify_identity
            iapp_cls.identity_policy()(iapp_get_identity_policy)
            iapp_cls.verify_identity()(iapp_verify_identity)
        else:
            iapp_cls.identity_policy()(get_identity_policy)
            iapp_cls.verify_identity()(verify_identity)

        iapp_cls.init_settings(settings)
        iapp_cls._raw_settings = settings
        iapp_cls.commit()

    app.identity_policy()(get_identity_policy)
    app.verify_identity()(verify_identity)
    app.authnz_provider()(lambda: authnz_provider())
    app.init_settings(settings)
    app._raw_settings = settings

    if settings['application']['development_mode']:
        os.environ['MOREPATH_TEMPLATE_AUTO_RELOAD'] = "1"

    app.commit()

    if settings['worker']['enabled']:
        celery_settings = settings['worker']['celery_settings']
        app.celery.conf.update(**celery_settings)
    application = app()
    return application


@create_app.register(app=SQLApp)
def create_sqlapp(app, settings, scan=True, **kwargs):

    application = create_baseapp(app=app, settings=settings, scan=scan,
                                 **kwargs)

    application.initdb()
    return application


def run(app, settings, host='127.0.0.1', port=5000, ignore_cli=True):
    application = create_app(app, settings)
    morepath.run(application, host=host, port=port, ignore_cli=ignore_cli)



import abc


class IStorage(abc.ABCMeta):

    @abc.abstractmethod
    def search_users(self, filters):
        "return list of users"
        raise NotImplementedError

    @abc.abstractmethod
    def search_groups(self, filters):
        "return list of groups"
        raise NotImplementedError

    @abc.abstractmethod
    def get_user(self, username):
        """
        get specific user model

        return User object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_user(self, username, password, attrs):
        """
        create a user

        return User object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def validate(self, username, password):
        """
        validate user authentication

        return boolean
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_group(self, groupname, attrs=None):
        """
        create a group

        return Group object
        raises ValueError if group already exists
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_group(self, groupname):
        """
        get group by name

        return Group object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_members(self, groupname):
        """
        get list of members of group

        return list of User objects        
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add_group_members(self, groupname, usernames):
        """
        add list of usernames into group
        ignore if already exists in group

        return None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def remove_group_members(self, groupname, usernames):
        """
        remove list of usernames from group
        ignore if doesnt exists in group

        return None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_group_user_roles(self, groupname, username):
        """
        get list of roles mapping in a group for user

        return list of role string
        """
        raise NotImplementedError

    @abc.abstractmethod
    def grant_group_user_role(self, groupname, username, rolename):
        """
        grant role in group to username
        ignore if role already granted

        return None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def revoke_group_user_role(self, groupname, username, rolename):
        """
        revoke role in group from username
        ignore if role doesn't exist

        return None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_user_groups(self, username):
        """
        search for group for specified user

        return a list of Group objects
        """
        raise NotImplementedError



from morpfw.crud.storage.sqlstorage import Base
import sqlalchemy as sa


class APIKey(Base):

    __tablename__ = 'authmanager_apikey'

    username = sa.Column(sa.String(256))
    label = sa.Column(sa.String(length=256))
    api_identity = sa.Column(sa.String(length=40))
    api_secret = sa.Column(sa.String(length=40))



from morpfw.crud import Schema, Collection, Model
from .schema import APIKeySchema
import jsl
from ..app import App
from .storage import APIKey
from uuid import uuid4
import rulez


class APIKeyModel(Model):
    schema = APIKeySchema


class APIKeyCollection(Collection):
    schema = APIKeySchema

    def search(self, query=None, *args, **kwargs):
        if kwargs.get('secure', True):
            if query:
                rulez.and_(
                    rulez.field['username'] == self.request.identity.userid,
                    query)
            else:
                query = rulez.field['username'] == self.request.identity.userid
        return super(APIKeyCollection, self).search(query, *args, **kwargs)



from morpfw.crud.storage.sqlstorage import Base
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson


class User(Base):

    __tablename__ = 'authmanager_users'

    username = sa.Column(sa.String(length=256))
    email = sa.Column(sa.String)
    password = sa.Column(sa.String(length=1024))
    attrs = sa.Column(sajson.JSONField)
    state = sa.Column(sa.String(length=64))



import morpfw
import jsonobject
import sqlalchemy as sa

class App(morpfw.SQLApp):
   pass

class PageSchema(morpfw.Schema):
   body = jsonobject.StringProperty(required=False)

class PageCollection(morpfw.Collection):
   schema = PageSchema

class PageModel(morpfw.Model):
   schema = PageSchema

class Page(morpfw.SQLBase):
   __tablename__ = 'test_page'

   body = sa.Column(sa.String(length=1024))

class PageStorage(morpfw.SQLStorage):
   model = PageModel
   orm_model = Page

@App.path(model=PageCollection, path='pages')
def get_collection(request):
   storage = PageStorage(request)
   return PageCollection(request, storage)

@App.path(model=PageModel, path='pages/{identifier}')
def get_model(request, identifier):
   collection = get_collection(request)
   return collection.get(identifier)

if __name__ == '__main__':
    morpfw.run(App, {})



#
from . import crud
from .crud import Collection, Model, Adapter, Schema
from .crud import StateMachine
from .crud.storage.sqlstorage import SQLStorage
from .crud.storage.elasticsearchstorage import ElasticSearchStorage
from .sql import Base as SQLBase
from . import auth as authmanager
from .app import SQLApp
from .main import create_app, run
from .app import create_admin
from .crud import signals as crudsignals
from .util import get_group, get_user



from .auth.app import App
from .auth import UserCollection
from .auth import GroupCollection
from .auth import permission as authperm
from .crud import permission as jslperm


def _has_admin_role(app, identity, context, permission):
    return app.has_role(context.request, 'administrator')


App.permission_rule(model=UserCollection,
                    permission=authperm.Register)(_has_admin_role)
App.permission_rule(model=UserCollection,
                    permission=jslperm.All)(_has_admin_role)
App.permission_rule(model=GroupCollection,
                    permission=jslperm.All)(_has_admin_role)



from morpfw.crud.storage.sqlstorage import Base
import sqlalchemy as sa
from sqlalchemy.orm import relationship
import sqlalchemy.orm as saorm
import sqlalchemy_jsonfield as sajson


class Group(Base):

    __tablename__ = 'authmanager_groups'

    groupname = sa.Column(sa.String(length=256))
    attrs = sa.Column(sajson.JSONField)
    memberships = relationship('Membership', cascade='all')


class Membership(Base):

    __tablename__ = 'authmanager_membership'

    group_id = sa.Column(sa.ForeignKey(
        'authmanager_groups.uuid'))
    user_id = sa.Column(sa.ForeignKey(
        'authmanager_users.uuid'))
    created = sa.Column(sa.DateTime)
    sa.UniqueConstraint('group_id', 'user_id')

    roles_assignment = relationship('RoleAssignment', cascade='all')


class RoleAssignment(Base):

    __tablename__ = 'authmanager_roleassignment'

    membership_id = sa.Column(sa.ForeignKey('authmanager_membership.uuid'))
    rolename = sa.Column(sa.String)
    created = sa.Column(sa.DateTime)
    sa.UniqueConstraint('membership_id', 'role_id')



from ..app import App
from .base import Provider
from ..types import datestr
from ..storage.memorystorage import MemoryStorage
from ..model import Model
import jsonobject
import datetime

_MARKER = []


class JsonObjectProvider(Provider):

    def __init__(self, schema, data, storage):
        self.schema = schema
        self.data = data
        self.storage = storage
        self.changed = False

    def __getitem__(self, key):
        if isinstance(self.schema.properties()[key], jsonobject.DateTimeProperty):
            value = self.data[key]
            assert isinstance(value, datetime.datetime) or value is None
            return value
        return self.data[key]

    def __setitem__(self, key, value):
        if isinstance(self.schema.properties()[key], jsonobject.DateTimeProperty):
            assert isinstance(value, datetime.datetime) or value is None

        self.data[key] = value
        self.changed = True

    def __delitem__(self, key):
        del self.data[key]
        self.changed = True

    def get(self, key, default=_MARKER):
        if default is _MARKER:
            return self.data[key]
        return self.data[key]

    def set(self, key, value):
        self.data[key] = value
        self.changed = True

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def to_json(self):
        return self.data.to_json()


@App.dataprovider(schema=jsonobject.JsonObject,
                  obj=jsonobject.JsonObject,
                  storage=MemoryStorage)
def get_dataprovider(model, obj, storage):
    return JsonObjectProvider(model, obj, storage)


@App.jsonprovider(obj=JsonObjectProvider)
def get_jsonprovider(obj):
    return obj.data.to_json()



import typing
from webob.dec import wsgify
from webob import exc
from webob.response import Response
from webob.static import FileIter
import os
BLOCK_SIZE = 1 << 16


class Blob(object):
    def __init__(self, uuid, filename, mimetype=None, size=None, encoding=None):
        self.uuid = uuid
        self.size = size
        self.filename = filename
        self.mimetype = mimetype
        self.encoding = encoding

    def open(self) -> typing.BinaryIO:
        raise NotImplementedError

    def get_size(self) -> int:
        return self.size

    @wsgify
    def __call__(self, req):
        if req.method not in ('GET', 'HEAD'):
            return exc.HTTPMethodNotAllowed("You cannot %s a file" %
                                            req.method)

        try:
            file = self.open()
        except (IOError, OSError) as e:
            msg = "You are not permitted to view this file (%s)" % e
            return exc.HTTPForbidden(msg)

        if 'wsgi.file_wrapper' in req.environ:
            app_iter = req.environ['wsgi.file_wrapper'](file, BLOCK_SIZE)
        else:
            app_iter = FileIter(file)

        return Response(
            app_iter=app_iter,
            content_length=self.get_size(),
            content_type=self.mimetype,
            content_encoding=self.encoding
            # @@ etag
        ).conditional_response_app


class BlobStorage(object):

    def get(self, uuid: str) -> Blob:
        raise NotImplementedError

    def put(self, fileobj: typing.BinaryIO,
            filename: str, mimetype: typing.Optional[str] = None,
            size: typing.Optional[int] = None,
            encoding: typing.Optional[str] = None) -> Blob:
        raise NotImplementedError

    def delete(self, uuid: str):
        raise NotImplementedError



import yaml
import os
from .test_auth import _test_authentication, get_client
from more.jwtauth import JWTIdentityPolicy
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from morpfw.app import SQLApp
from morpfw.auth.app import App


class SQLStorageAuthApp(App, SQLApp):
    pass


class SQLStorageApp(SQLApp):
    pass


@SQLStorageApp.mount(app=SQLStorageAuthApp, path='/auth')
def mount_app(app):
    return SQLStorageAuthApp()


def test_authentication_sqlstorage(pgsql_db):
    with open(os.path.join(os.path.dirname(__file__),
                           'settings-sqlalchemy.yml')) as f:
        settings = yaml.load(f)

    c = get_client(SQLStorageApp, settings)

    _test_authentication(c)



import morpfw
from morpfw import sql as morpsql
from morpfw.crud.schema import Schema
import sqlalchemy as sa
import jsl
from .common import get_client
import jsonobject


class App(morpfw.SQLApp):
    pass


class Page(morpsql.Base):
    __tablename__ = 'test_page'

    title = sa.Column(sa.String(length=1024))
    body = sa.Column(sa.Text)


class PageSchema(Schema):
    title = jsonobject.StringProperty()
    body = jsonobject.StringProperty()


@App.identifierfields(schema=PageSchema)
def page_schema_identifier(schema):
    return ['uuid']


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path='/')
def get_pagecollection(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='/{identifier}')
def get_page(request, identifier):
    storage = PageStorage(request)
    return storage.get(identifier)


def test_morp_framework(pgsql_db):
    c = get_client(App)

    r = c.get('/')
    assert len(r.json['schema']['properties']) == 8

    r = c.post_json(
        '/', {'title': 'Hello world', 'body': 'Lorem ipsum'})

    assert r.json['links'][0]['href'].startswith('http://localhost/')
    assert r.json['data']['title'] == 'Hello world'
    assert r.json['data']['body'] == 'Lorem ipsum'

    page_url = r.json['links'][0]['href']
    r = c.get(page_url)

    assert r.json['data']['title'] == 'Hello world'

    delete_link = r.json['links'][2]
    assert delete_link['method'] == 'DELETE'

    r = c.delete(delete_link['href'])

    r = c.get(page_url, expect_errors=True)

    assert r.status_code == 404



from ..app import App
from morpfw.crud.storage.memorystorage import MemoryStorage
from ..user.model import User, UserModel, UserSchema
from ..group.model import Group, GroupModel, GroupSchema
from ..apikey.model import APIKey, APIKeyModel, APIKeySchema
from .interfaces import IStorage
from morpfw.crud import errors as cruderrors
from .. import exc
import rulez

DB: dict = {
    'users': {},
    'groups': {},
    'rolemap': {}
}


class UserMemoryStorage(MemoryStorage):
    model = UserModel

    def get_by_email(self, email):
        users = self.search(rulez.field['email'] == email)
        if not users:
            return None
        return users[0]

    def change_password(self, username, new_password):
        user = self.get(username)
        user.data['password'] = new_password

    def get_user_groups(self, username):
        storage = self.request.app.root.get_authnz_provider().get_authn_storage(
            self.request, GroupSchema)
        return storage.get_user_groups(username)

    def validate(self, username, password):
        user = self.get(username)
        return user.data['password'] == password


class APIKeyMemoryStorage(MemoryStorage):
    model = APIKeyModel


class GroupMemoryStorage(MemoryStorage):
    model = GroupModel

    def get_user_groups(self, username):
        res = []
        for gid, group in self.datastore.items():
            if username in group.data['attrs'].get('members'):
                res.append(group)
        return res

    def get_members(self, groupname):
        group = self.get(groupname)
        userstorage = self.request.app.root.get_authnz_provider().get_authn_storage(
            self.request, UserSchema)
        res = []
        group.data.setdefault('attrs', {})
        attrs = group.data['attrs']
        attrs.setdefault('members', [])
        for m in group.data['attrs'].get('members'):
            res.append(userstorage.get(m))
        return res

    def add_group_members(self, groupname, usernames):
        group = self.get(groupname)
        group.data.setdefault('attrs', {})
        attrs = group.data['attrs']
        attrs.setdefault('members', [])
        for u in usernames:
            if u not in attrs['members']:
                attrs['members'].append(u)
        group.data['attrs'] = attrs

    def remove_group_members(self, groupname, usernames):
        group = self.get(groupname)
        group.data.setdefault('attrs', {})
        attrs = group.data['attrs']
        attrs.setdefault('members', [])
        for u in usernames:
            if u in attrs['members']:
                attrs['members'].remove(u)
        group.data['attrs'] = attrs

    def get_group_user_roles(self, groupname, username):
        rolemap = DB['rolemap']
        rolemap.setdefault(groupname, {})
        rolemap[groupname].setdefault(username, [])
        return rolemap[groupname][username]

    def grant_group_user_role(self, groupname, username, rolename):
        rolemap = DB['rolemap']
        rolemap.setdefault(groupname, {})
        rolemap[groupname].setdefault(username, [])
        if rolename not in rolemap[groupname][username]:
            rolemap[groupname][username].append(rolename)

    def revoke_group_user_role(self, groupname, username, rolename):
        rolemap = DB['rolemap']
        rolemap.setdefault(groupname, {})
        rolemap[groupname].setdefault(username, [])
        if rolename in rolemap[groupname][username]:
            rolemap[groupname][username].remove(rolename)


SINGLETON: dict = {}


def userstorage_factory(request, *args, **kwargs):
    return UserMemoryStorage(request)


@App.authn_storage('memorystorage', UserSchema)
def get_userstorage_factory(name, schema):
    return userstorage_factory


def apikeystorage_factory(request, *args, **kwargs):
    return APIKeyMemoryStorage(request)


@App.authn_storage('memorystorage', APIKeySchema)
def get_apikeystorage_factory(name, schema):
    return apikeystorage_factory


def groupstorage_factory(request, *args, **kwargs):
    return GroupMemoryStorage(request)


@App.authn_storage('memorystorage', GroupSchema)
def get_groupstorage_factory(name, schema):
    return groupstorage_factory



from jsonschema.validators import Draft4Validator
from jsonschema import validate, ValidationError
from .util import jsl_nullable
from .util import jsonobject_to_jsl
from .const import SEPARATOR
from . import permission
from . import signals
from .log import logger
from rulez import validate_condition
from morepath import reify
from DateTime import DateTime
from uuid import uuid4
from transitions import Machine
import copy
from .errors import StateUpdateProhibitedError, AlreadyExistsError, BlobStorageNotImplementedError
import jsonobject


ALLOWED_SEARCH_OPERATORS = [
    'and', 'or', '==', 'in',
    '~', '!=', '>', '<', '>=',
    '<='
]


class Collection(object):

    create_view_enabled = True
    search_view_enabled = True
    aggregate_view_enabled = True

    exist_exc = AlreadyExistsError

    @property
    def schema(self):
        raise NotImplementedError

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __init__(self, request, storage, data=None):
        self.request = request
        self.app = request.app
        self.storage = storage
        self.data = None
        if data:
            self.data = request.app.get_dataprovider(self.schema, data,
                                                     self.storage)

    def search(self, query=None, offset=0, limit=None, order_by=None,
               secure=False):
        objs = self._search(query, offset, limit, order_by, secure)
        if secure and limit:
            nextpage = {'query': query, 'offset': offset +
                        limit, 'limit': limit, 'order_by': order_by}
            while len(objs) < limit:
                nextobjs = self._search(secure=True, **nextpage)
                if len(nextobjs) == 0:
                    return list(objs[:limit])
                nextpage['offset'] = nextpage['offset'] + limit
                objs = objs + nextobjs
        return objs

    def _search(self, query=None, offset=0, limit=None, order_by=None,
                secure=False):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)
        objs = self.storage.search(
            query, offset=offset, limit=limit, order_by=order_by)

        if secure:
            objs = list([obj for obj in objs if self.request.app.permits(
                self.request, obj, permission.View)])
        return list(objs)

    def aggregate(self, query=None, group=None, order_by=None):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)
        objs = self.storage.aggregate(query, group=group, order_by=order_by)
        return list(objs)

    def create(self, data):
        identifier = self.app.get_default_identifier(
            self.schema, data, self.request)
        if identifier and self.get(identifier):
            raise self.exist_exc(identifier)
        obj = self._create(data)
        obj.set_initial_state()
        self.request.app.signal_publish(self.request,
                                        obj, signals.OBJECT_CREATED)
        return obj

    def _create(self, data):
        data = self.storage.set_schema_defaults(data)
        return self.storage.create(data)

    def get(self, identifier):
        if isinstance(identifier, list) or isinstance(identifier, tuple):
            identifier = self.request.app.join_identifier(*identifier)
        return self.storage.get(identifier)

    def get_by_uuid(self, uuid):
        return self.storage.get_by_uuid(uuid)

    def json(self):
        return {
            'schema': jsonobject_to_jsl(self.schema).get_schema(ordered=True),
            'links': self.links()
        }

    def links(self):
        request = self.request
        links = [{'rel': 'create',
                  'href': request.link(self),
                  'method': 'POST'},
                 {'rel': 'search',
                  'href': request.link(self, '+search')}]
        links += self._links()
        return links

    def _links(self):
        return []


class Model(object):

    linkable = True
    update_view_enabled = True
    delete_view_enabled = True

    blobstorage_field = 'blobs'
    blob_fields = []

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __dict__(self):
        return self.data.as_dict()

    @property
    def statemachine_view_enabled(self):
        if self.state_machine():
            return True
        return False

    @property
    def schema(self):
        raise NotImplementedError

    @property
    def identifier(self):
        if self._cached_identifier:
            return self._cached_identifier
        res = []
        for f in self.app.get_identifierfields(self.schema):
            d = self.data.get(f)
            if d is not None:
                d = str(d)
            res.append(d)
        if None in res:
            identifier = self.app.get_default_identifier(
                self.schema, self.data, self.request)
            if identifier is None:
                return None
            if isinstance(identifier, list) or isinstance(identifier, tuple):
                identifier = self.app.join_identifier(*identifier)
            self.storage.set_identifier(self.data, identifier)
            self._cached_identifier = identifier
            return identifier
        separator = self.request.app.get_compositekey_separator()
        identifier = separator.join(res)
        self._cached_identifier = identifier
        return identifier

    @property
    def uuid(self):
        uuid_field = self.app.get_uuidfield(self.schema)
        return self.data[uuid_field]

    def __init__(self, request, storage, data):
        self.request = request
        self.storage = storage
        self.app = request.app
        self.data = request.app.get_dataprovider(self.schema, data,
                                                 self.storage)
        self._cached_identifier = None

    def update(self, newdata):
        if 'state' in newdata:
            raise StateUpdateProhibitedError()

        data = self._raw_json()
        data.update(newdata)
        schema = jsonobject_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        validate(data, schema)
        self.storage.update(self.identifier, data)
        self.request.app.signal_publish(
            self.request, self, signals.OBJECT_UPDATED)

    def delete(self):
        self.request.app.signal_publish(
            self.request, self, signals.OBJECT_TOBEDELETED)
        self.storage.delete(self.identifier, model=self)

    def save(self):
        if self.data.changed:
            data = self._raw_json()
            schema = jsonobject_to_jsl(
                self.schema, nullable=True).get_schema(ordered=True)
            validate(data, schema)
            self.storage.update(self.identifier, data)

    def _raw_json(self):
        schema = jsonobject_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        jsondata = self.app.get_jsonprovider(self.data)
        jsondata = self.rules_adapter().transform_json(copy.deepcopy(jsondata))
        try:
            validate(jsondata, schema)
        except ValidationError as e:
            logger.warn('%s(%s) : %s' % (self.schema.__name__,
                                         '/'.join(list(e.path)), e.message))
        return jsondata

    def _json(self):
        return self._raw_json()

    def json(self):
        if self.linkable:
            return {
                'data': self._json(),
                'links': self.links()
            }
        return self._json()

    def links(self):
        links = []
        links.append({
            'rel': 'self',
            'href': self.request.link(self)
        })
        if self.update_view_enabled:
            links.append({
                'rel': 'update',
                'href': self.request.link(self),
                'method': 'PATCH'
            })
        if self.delete_view_enabled:
            links.append({
                'rel': 'delete',
                'href': self.request.link(self),
                'method': 'DELETE'
            })
        if self.statemachine_view_enabled:
            links.append({
                'rel': 'statemachine',
                'href': self.request.link(self, '+statemachine'),
                'method': 'POST'
            })
        links += self._links()
        links += self.rules_adapter().links()
        return links

    def _links(self):
        return []

    def rules_adapter(self):
        return self.app.get_rulesadapter(self)

    def state_machine(self):
        if self.app.get_statemachine.by_args(self).all_matches:
            return self.app.get_statemachine(self)
        return None

    def set_initial_state(self):
        self.state_machine()

    def _blob_guard(self, field):
        if self.blobstorage_field not in self.schema._properties_by_attr.keys():
            raise BlobStorageNotImplementedError(
                'Object does not implement blobs store')
        if field not in self.blob_fields:
            raise BlobStorageNotImplementedError(
                'Field %s not allowed for blobstorage' % field)

    def put_blob(self, field, fileobj, filename, mimetype=None, size=None, encoding=None):
        self._blob_guard(field)
        blob_data = self.data[self.blobstorage_field]
        existing = blob_data.get(field, None)
        blob = self.storage.put_blob(
            fileobj, filename, mimetype, size, encoding)
        blob_data[field] = blob.uuid
        self.update({self.blobstorage_field: blob_data})
        if existing:
            self.delete_blob(existing)
        return blob

    def get_blob(self, field):
        uuid = self.data[self.blobstorage_field][field]
        blob = self.storage.get_blob(uuid)
        return blob

    def delete_blob(self, field):
        uuid = self.data[self.blobstorage_field][field]
        self.storage.delete_blob(uuid)


class StateMachine(object):

    @property
    def states(self):
        raise NotImplementedError

    @property
    def transitions(self):
        raise NotImplementedError

    def __init__(self, context):
        self._context = context
        self._request = context.request
        self._app = context.request.app
        initial = self.state or self.states[0]
        self._machine = Machine(
            model=self, transitions=self.transitions,
            states=self.states, initial=initial)

    @property
    def state(self):
        try:
            return self._context.data['state']
        except KeyError:
            return None

    @state.setter
    def state(self, val):
        self._context.data['state'] = val

    def get_triggers(self):
        return [i for i in self._machine.get_triggers(self.state) if not i.startswith('to_')]



import jsonobject
from morpfw.crud import Schema


class APIKeySchema(Schema):

    username = jsonobject.StringProperty()
    label = jsonobject.StringProperty()
    api_identity = jsonobject.StringProperty()
    api_secret = jsonobject.StringProperty()



from .test_auth import _test_authentication, get_client
from more.jwtauth import JWTIdentityPolicy
from morpfw.app import BaseApp
from morpfw.auth.app import App
import os
import yaml


class MemoryStorageAuthApp(App, BaseApp):
    pass


class MemoryStorageApp(BaseApp):
    pass


@MemoryStorageApp.mount(app=MemoryStorageAuthApp, path='/auth')
def mount_app(app):
    return MemoryStorageAuthApp()


def test_authentication_memorystorage():
    with open(os.path.join(os.path.dirname(__file__),
                           'settings-memorystorage.yml')) as f:
        settings = yaml.load(f)

    c = get_client(MemoryStorageApp, settings)
    _test_authentication(c)



import jsonobject
from uuid import uuid4
from .app import App

class Schema(jsonobject.JsonObject):

    uuid = jsonobject.StringProperty(required=False, 
        default=lambda: uuid4().hex)
    creator = jsonobject.StringProperty(required=False)
    created = jsonobject.DateTimeProperty(required=False)
    modified = jsonobject.DateTimeProperty(required=False)
    state = jsonobject.StringProperty(required=False)
    deleted = jsonobject.DateTimeProperty(required=False)


@App.identifierfields(schema=Schema)
def identifierfields(schema):
    return ['uuid']

@App.default_identifier(schema=Schema)
def default_identifier(schema, obj, request):
    fields = request.app.get_identifierfields(schema)
    res = []
    for f in fields:
        if f == 'uuid':
            res.append(uuid4().hex)
        else:
            res.append(obj[f])
    return res





import jsonobject
from morpfw import Schema
from ..app import App


class GroupSchema(Schema):
    groupname = jsonobject.StringProperty(
        required=True)  # , pattern=NAME_PATTERN)
    members = jsonobject.ListProperty(str, required=False)
    attrs = jsonobject.DictProperty(required=False)
    created = jsonobject.StringProperty(required=False)
    modified = jsonobject.StringProperty(required=False)


@App.identifierfields(schema=GroupSchema)
def group_identifierfields(schema):
    return ['groupname']


class MemberSchema(Schema):
    users = jsonobject.ListProperty(str, required=True)



import dectate
import reg


class StorageAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name, schema):
        self.name = name
        self.schema = schema

    def identifier(self, app_class):
        return str((self.name, self.schema))

    def perform(self, obj, app_class):
        app_class._get_authn_storage.register(
            reg.methodify(obj), name=self.name, schema=self.schema)



from morpfw.crud import Collection, Model, StateMachine
from morpfw.crud import errors as cruderrors
from .storage import User
from ..app import App
import jsl
from .. import exc
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from ..model import NAME_PATTERN, EMAIL_PATTERN
from morpfw.crud import signals as crudsignal
from morpfw.crud import errors as cruderrors
from morpfw.crud.schema import Schema
from morpfw.crud.rulesadapter import Adapter
from morpfw.crud.validator import regex_validator
from ..group.model import GroupCollection, GroupSchema
from uuid import uuid4
import re
import jsonobject
from .schema import RegistrationSchema, UserSchema, LoginSchema
from ..exc import UserExistsError


class UserCollection(Collection):
    schema = UserSchema
    exist_exc = UserExistsError

    def authenticate(self, username, password):
        if re.match(EMAIL_PATTERN, username):
            user = self.storage.get_by_email(username)
            if user is None:
                return False
            return user.validate(password)

        user = self.storage.get(username)
        if user is None:
            return False
        return user.validate(password)

    def _create(self, data):
        data['nonce'] = uuid4().hex
        return super(UserCollection, self)._create(data)


class UserModel(Model):

    schema = UserSchema

    def change_password(self, password, new_password):
        if not self.app.has_role(self.request, 'administrator'):
            if not self.validate(password, check_state=False):
                raise exc.InvalidPasswordError(self.data['username'])
        self.storage.change_password(self.data['username'], new_password)

    def validate(self, password, check_state=True):
        if check_state and self.data['state'] != 'active':
            return False
        return self.storage.validate(self.data['username'], password)

    def groups(self):
        return self.storage.get_user_groups(self.data['username'])

    def group_roles(self):
        group_roles = {}
        for g in self.groups():
            group_roles[g.data['groupname']] = g.get_member_roles(
                self.data['username'])
        return group_roles


class UserStateMachine(StateMachine):

    states = ['active', 'inactive', 'deleted']
    transitions = [
        {'trigger': 'activate', 'source': 'inactive', 'dest': 'active'},
        {'trigger': 'deactivate', 'source': 'active', 'dest': 'inactive'},
        {'trigger': 'delete', 'source': [
            'active', 'inactive'], 'dest': 'deleted'}
    ]

    def on_enter_deleted(self):
        self._context.delete()


@App.statemachine(model=UserModel)
def userstatemachine(context):
    return UserStateMachine(context)


class UserRulesAdapter(Adapter):

    def transform_json(self, data):
        data = data.copy()
        for f in ['password', 'password_validate']:
            if f in data.keys():
                del data[f]
        data['groups'] = [g.identifier for g in self.context.groups()]
        return data


@App.rulesadapter(model=UserModel)
def get_rulesadapter(obj):
    return UserRulesAdapter(obj)


@App.subscribe(signal=crudsignal.OBJECT_CREATED, model=UserModel)
def add_user_to_default_group(app, request, obj, signal):
    request = obj.request
    storage = app.get_authn_storage(request, GroupSchema)
    g = storage.get('__default__')
    if g is None:
        gcol = GroupCollection(request, storage)
        g = gcol.create({'groupname': '__default__'})
    g.add_members([obj.data['username']])



from .app import App
from .model import Collection, Model
from . import permission


@App.permission_rule(model=Model, permission=permission.All)
def model_permissions(identity, model, permission):
    return True


@App.permission_rule(model=Collection, permission=permission.All)
def collection_permissions(identity, model, permission):
    return True
